/*
 *	Copyright (C) 2013-2017 Jan Bobrowski <jb@torinak.com>
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License
 *	version 2 as published by the Free Software Foundation.
 */

#include <stdlib.h>
#include "ttf2woff.h"

#include "zopfli/zlib_container.c"
#include "zopfli/deflate.c"
#include "zopfli/lz77.c"
#include "zopfli/blocksplitter.c"
#include "zopfli/squeeze.c"
#include "zopfli/hash.c"
#include "zopfli/cache.c"
#include "zopfli/tree.c"
#include "zopfli/util.c"
#include "zopfli/katajainen.c"

#define adler32 zlib_adler32
#include <zlib.h>

char *copression_by = "zopfli";

int zlib_compress(struct buf *out, struct buf inp)
{
	ZopfliOptions options = {.numiterations = 20};
	u8 *buffer = 0;
	size_t size = 0;

	ZopfliZlibCompress(&options, inp.ptr, inp.len, &buffer, &size);
	if(size >= inp.len) {
		free(buffer);
		return 0;
	}

#if 1
	{
		/* Trust, but verify */
		uLong tmpl = inp.len;
		Bytef *tmpb = my_alloc(tmpl);
		int v = uncompress(tmpb, &tmpl, buffer, size);
		if(v != Z_OK || tmpl != inp.len)
			errx(3, "Zopfli error");
		my_free(tmpb);
	}
#endif

	out->ptr = buffer;
	out->len = size;
	return 1;
}

#define ZLIB_DECOMPRESS_ONLY
#include "comp-zlib.c"
