/*
 *	Copyright (C) 2013-2017 Jan Bobrowski <jb@torinak.com>
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License
 *	version 2 as published by the Free Software Foundation.
 */

#include <string.h>
#include <stdlib.h>
#include <zlib.h>
#include "ttf2woff.h"

static struct buf get_or_inflate(u8 *p, size_t len, size_t orig_len)
{
	struct buf inp;
	struct buf out;

	if(len == orig_len) {
		out.ptr = p;
		out.len = len;
		return out;
	}

	inp.ptr = p;
	inp.len = (int)len;
	out.len = orig_len;
	out.ptr = my_alloc(orig_len);

	assert(inp.len == len);

	zlib_decompress(out, inp);

	return out;
}

void read_woff(struct ttf *ttf, u8 *data, size_t length)
{
	u8 *p;
	int i;

	if(length <= 44 + 20) ERR_TRUNCATED;

	ttf->flavor = g32(data+4);

	ttf->ntables = g16(data+12);
	if(!ttf->ntables) ERR_FONT;

	if(length < g32(data+8)) ERR_TRUNCATED;
	length = g32(data+8);

	ttf->font_version = g32(data+20);

	{
		u32 len = g32(data+28);
		ttf->woff_meta.len = 0;
		if(len) {
			u32 ofs = g32(data+24);
			if(!sum_in_range(ofs, len, length))
				ERR_FONT;
			ttf->woff_meta = get_or_inflate(data + ofs, len, g32(data+32));
		}
	}

	ttf->woff_priv.len = g32(data+40);
	ttf->woff_priv.ptr = ttf->woff_priv.len ? data+g32(data+36) : 0;

	alloc_tables(ttf);

	p = data + 44;
	for(i=0; i<ttf->ntables; i++) {
		struct table *t = &ttf->tables[i];
		u32 ofs = g32(p+4);
		u32 len = g32(p+8);
		if(!sum_in_range(ofs, len, length))
			ERR_FONT;
		t->tag = g32(p);
		t->csum = g32(p+16);
		t->pos = ofs;
		t->free_buf = 1;
		t->buf = get_or_inflate(data + ofs, len, g32(p+12));
		name_table(t);
		p += 20;
	}
}
